import * as Koa from 'koa';
import { Server as HTTPServer } from 'http';
import { Server as HTTPSServer } from 'https';
import RouteMapper from 'route-mapper';

interface IAppConstructor {
  app: string;
  label: string;
  controllers: IControllerConstructor[];
  tasks?: ITaskConstructor[];
  models?: IModelConstructor[];
  meta?: any;
  [ key: string ]: any;
  new (...args: any[]): App;
}

interface IControllerConstructor {
  app?: string;
  controller: string;
  new(): Controller;
}

interface IMiddleware {
  (ctx: any, next?: Function): Promise<any>;
}

interface IModelConstructor {
  app?: string;
  new (...args: any[]): Model;
}

interface IServers {
  http: HTTPServer | undefined;
  https: HTTPSServer | undefined;
}

interface ITaskConstructor {
  app: string;
  command: string;
  new (...args: any[]): Task;
}

interface IOverlandSettings {
  apps: IAppConstructor[];
  keys?: string[];
  middleware: IMiddleware[];
  ports: {
    http: number;
    https?: number;
    ws?: number;
  };
  ssl?: {
    key: Buffer;
    cert: Buffer;
  };
  [key: string]: any;
}

export default class Overland extends Koa {
  public settings: IOverlandSettings;
  public apps: Map<string, App>;
  public controllers: Map<string, IControllerConstructor>;
  public engine: any;
  public models: Map<string, IModelConstructor>;
  public keys: string[];
  public helpers: { [ key: string]: Function };
  public createRenderContext(ctx: any, data: any): any;
  public serve(): IServers;
  public use(middleware: IMiddleware): Overland;
  public boot(): Promise<this>;
  constructor(settings: IOverlandSettings);
}

export abstract class App {
  public static app: string;
  public static routes: Router;
  public static controllers: IControllerConstructor[];
  public static models: IModelConstructor[];
  public static label: string;
  public static meta: any;
  public static engine: any;
  public static views: string;

  private static _app: string;
  private static _router: Router;
  private static _controllers: IControllerConstructor[];
  private static _models: IModelConstructor[];

  public app: string;
  public engine: Function;
  public label: string;
  public views: string;
  public meta: any;
  public controllers: IControllerConstructor[];
  public models: IModelConstructor[];
  public init(app: Overland): Promise<any>
}

/**
 * Abstract class; subclasses should provide async methods handlers for HTTP requests.
 */
export abstract class Controller {
  public static app: string;
  public static controller: string;
  private static _controller: string | null;
  public name: string;
  public action: string;
  public app: string;
}

export class Model {
  public static app: string;
  public save(): Promise<this>;
  constructor(opts: any);
}

export class Task {
  public static app: string;
  public static command: string;
  private static _command: string;
  private static _app: string;
  public overland: Overland;
  public action(...args): Promise<any>;
  constructor(overland: Overland);
}

export class View {
  public data: any;
  public methods(): IterableIterator<string[]>;
  constructor(data?: any);
}

export class Router extends RouteMapper {
  public app: string;
  public routes: any[];
  public channel(path: string, options?: any): this;
  public use(path: string | Router, router?: Router): this
}